@@ -2,12 +2,13 @@ package Plack::Middleware::CSRFBlock;
 use parent qw(Plack::Middleware);
 use strict;
 use warnings;
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 use HTML::Parser;
+use Plack::TempBuffer;
 use Plack::Util::Accessor qw(
     parameter_name token_length session_key blocked onetime
-    _param_re_urlenc _param_re_formdata _token_generator
+    _param_re _token_generator
 );
 
 sub prepare_app {
@@ -20,18 +21,19 @@ sub prepare_app {
     my $parameter_name = $self->parameter_name;
     my $token_length = $self->token_length;
 
-    $self->_param_re_urlenc(qr/
-        (?:^|&)
-        $parameter_name=([0-9a-f]{$token_length})
-        (?:&|$)
-    /x);
-
-    $self->_param_re_formdata(qr/
-        ; ?name="?$parameter_name"?(?:;[^\x0d]*)?\x0d\x0a
-        (?:[^\x0d]+\x0d\x0a)*
-        \x0d\x0a
-        ([0-9a-f]{$token_length})\x0d\x0a
-    /x);
+    $self->_param_re({
+        'application/x-www-form-urlencoded' => qr/
+            (?:^|&)
+            $parameter_name=([0-9a-f]{$token_length})
+            (?:&|$)
+        /x,
+        'multipart/form-data' => qr/
+            ; ?name="?$parameter_name"?(?:;[^\x0d]*)?\x0d\x0a
+            (?:[^\x0d]+\x0d\x0a)*
+            \x0d\x0a
+            ([0-9a-f]{$token_length})\x0d\x0a
+        /x,
+    });
 
     $self->_token_generator(sub {
         my $token = Digest::SHA1::sha1_hex(rand() . $$ . {} . time);
@@ -47,23 +49,17 @@ sub call {
         die "CSRFBlock needs Session.";
     }
 
-    my $token = $session->{$self->session_key};
-    my $parameter_name = $self->parameter_name;
-
     # input filter
-    if($env->{REQUEST_METHOD} eq 'POST' and not $token) {
-        return $self->token_not_found;
-    }
-    elsif($env->{REQUEST_METHOD} eq 'POST') {
-        my $ct = $env->{CONTENT_TYPE};
+    if($env->{REQUEST_METHOD} eq 'POST' and
+        ($env->{CONTENT_TYPE} eq 'application/x-www-form-urlencoded' or
+         $env->{CONTENT_TYPE} eq 'multipart/form-data')
+    ) {
+        my $token = $session->{$self->session_key}
+            or return $self->token_not_found;
+
         my $cl = $env->{CONTENT_LENGTH};
-        my $re = 
-              $ct eq 'application/x-www-form-urlencoded' ? $self->_param_re_urlenc
-            : $ct eq 'multipart/form-data'               ? $self->_param_re_formdata
-            : undef;
-
+        my $re = $self->_param_re->{$env->{CONTENT_TYPE}};
         my $input = $env->{'psgi.input'};
-
         my $buffer;
 
         if ($env->{'psgix.input.buffered'}) {
@@ -91,6 +87,7 @@ sub call {
                         last if not $buffer;
                     }
                     $buffer->print($buf);
+                    undef $buf;
                     $done = 1;
                 }
             }
@@ -100,13 +97,14 @@ sub call {
             }
         }
 
-        if(not $found) {
+        if($found) {
+            # clear token if onetime option is enabled.
+            delete $session->{$self->session_key} if $self->onetime;
+        }
+        else {
             return $self->token_not_found($env);
         }
 
-        # clear token if onetime option is enabled.
-        undef $token if $self->onetime;
-
         if($buffer) {
             $env->{'psgi.input'} = $buffer->rewind;
             $env->{'psgix.input.buffered'} = Plack::Util::TRUE;
@@ -116,10 +114,6 @@ sub call {
         }
     }
 
-    # generate token
-    if(not $token) {
-        $session->{$self->session_key} = $token = $self->_token_generator->();
-    }
 
     return $self->response_cb($self->app->($env), sub {
         my $res = shift;
@@ -129,6 +123,9 @@ sub call {
         }
 
         my @out;
+        my $http_host = exists $env->{HTTP_HOST} ? $env->{HTTP_HOST} : $env->{SERVER_NAME};
+        my $token = $session->{$self->session_key} ||= $self->_token_generator->();
+        my $parameter_name = $self->parameter_name;
 
         my $p = HTML::Parser->new(
             api_version => 3,
@@ -136,21 +133,23 @@ sub call {
                 my($tag, $attr, $text) = @_;
                 push @out, $text;
 
-                #TODO: should we exclude form that action to outside?
-                if(lc($tag) eq 'form') {
-                    no warnings 'uninitialized';
-                    if(lc($attr->{'method'}) eq 'post') {
-                        my $token = $session->{$self->session_key};
-                        # TODO: determine xhtml or html?
-                        push @out, qq{<input type="hidden" name="$parameter_name" value="$token" />};
-                    }
+                no warnings 'uninitialized';
+                if(
+                    lc($tag) ne 'form' or
+                    lc($attr->{'method'}) ne 'post' or
+                    ($attr->{'action'} =~ m{^https?://([^/])+/} and $1 ne $http_host)
+                ) {
+                    return;
                 }
+                # TODO: determine xhtml or html?
+
+                push @out, qq{<input type="hidden" name="$parameter_name" value="$token" />};
 
             }, "tagname, attr, text"],
             default_h => [\@out , '@{text}'],
         );
         my $done;
-            
+
         return sub {
             return if $done;
 
